###########################################################################
# Replication Script for Generating the Tables and Figures in the Paper
###########################################################################


###########################################################################
##################### Install all necessary packages ######################
###########################################################################
# R-version: 4.0.3

# install.packages("Ball")              # (recommended) version number: 1.3.10
# install.packages("ape")               # (recommended) version number: 5.4-1
# install.packages("fda")               # (recommended) version number: 5.1.9
# install.packages("HHG")               # (recommended) version number: 2.3.2
# install.packages("movMF")             # (recommended) version number: 0.2-5
# install.packages("energy")            # (recommended) version number: 1.7.7
# install.packages("NPCirc")            # (recommended) version number: 2.0.1
# install.packages("fda.usc")           # (recommended) version number: 2.0.2
# install.packages("ggplot2")           # (recommended) version number: 3.3.3
# install.packages("mvtnorm")           # (recommended) version number: 1.1-1
# install.packages("reshape2")          # (recommended) version number: 1.4.4
# install.packages("circular")          # (recommended) version number: 0.4-93
# install.packages("EnvStats")          # (recommended) version number: 2.4.0
# install.packages("CovTools")          # (recommended) version number: 0.5.3
# install.packages("snowfall")          # (recommended) version number: 1.84-6.1
# install.packages("treespace")         # (recommended) version number: 1.1.4.0
# install.packages("gridExtra")         # (recommended) version number: 2.3
# install.packages("multivariance")     # (recommended) version number: 2.4.0
# install.packages("microbenchmark")    # (recommended) version number: 1.4-7
# install.packages("clusterGeneration") # (recommended) version number: 1.3.7

rm(list = ls()); gc(reset = TRUE)
###########################################################################
############################# Quick examples ##############################
###########################################################################

######################## The first quick examples #########################
library("Ball")
set.seed(1)
x <- rnorm(50)
y <- rnorm(50, mean = 1)
bd.test(x = x, y = y)

######################## The second quick examples ########################
library("mvtnorm")
set.seed(1)
cov_mat <- matrix(0.3, 3, 3)
diag(cov_mat) <- 1
data_set <- rmvnorm(n = 100, mean = c(0, 0, 0), sigma = cov_mat)
data_set <- as.list(as.data.frame(data_set))
bcov.test(x = data_set)


###########################################################################
########################### Real world examples ###########################
###########################################################################

######################### Wind direction dataset ##########################
library("Ball")
data("speed.wind", package = "NPCirc")
index1 <- which(speed.wind[["Year"]] == 2007 & speed.wind[["Month"]] == 11 &
                  speed.wind[["Day"]] %in% 1:7)
index2 <- which(speed.wind[["Year"]] == 2008 & speed.wind[["Month"]] == 1 &
                  speed.wind[["Day"]] %in% 25:31)
d1 <- speed.wind[["Direction"]][index1]
d2 <- speed.wind[["Direction"]][index2]
d1 <- na.omit(d1)
d2 <- na.omit(d2)
size_vec <- c(length(d1), length(d2))
theta <- c(d1, d2) / 360
dat <- cbind(cos(theta), sin(theta))
dx <- nhdist(dat, method = "geodesic")
bd.test(dx, size = size_vec, distance = TRUE)

### Plot Figure 1
library("circular")
# pdf(file = "../Figures/wind_direction.pdf", height = 6, width = 8)
par(mfrow = c(1, 2), mar = c(0, 0, 0, 0))
plot(circular(c(d1), units = "degrees"), cex = 0.8, bin = 100, stack = TRUE,
     shrink = 1.3)
plot(circular(c(d2), units = "degrees"), cex = 0.8, bin = 100, stack = TRUE,
     shrink = 1.3)
# dev.off()

########################## Brain fMRI dataset ###########################
## Note: To reproduce the "niICBM.rda" file, 
## you should conduct the following steps:
## 1. Download ICBM_part1.tar, ICBM_part2.tar, and ICBM_part3.tar from
## \url{https://drive.google.com/drive/folders/1SIVvZ7uLpPAgnK_7JmohyxMB04k9oXkR?usp=sharing}
## or \url{https://www.nitrc.org/frs/?group_id=296}
## 2. Run commands listed in file fmri.sh
## Platform: Win 10, 64-bit
## Python Environment: 
## Anaconda version: 2020.02, 64-bit; Python version: 3.6.12

library("Ball")
library("CovTools")
load("niICBM.rda")
dx <- CovDist(niICBM[["spd"]])
dy <- dist(niICBM[["covariate"]][["V3"]])
bcov.test(x = dx, y = dy, distance = TRUE, weight = "prob")


###########################################################################
############################ Numerical studies ############################
###########################################################################

library("snowfall")

############################## K-sample test ##############################

rm(list = ls()); gc(reset = TRUE)

data_generator <- function(seed = 1, case = "example1", 
                           size = c(10, 10, 10, 10)) {
  set.seed(seed)
  n <- sum(size)
  if (case == "example1") {
    x <- movMF::rmovMF(n = sum(size), 30 * c(1, 0, 0))
    x <- Ball::nhdist(x, method = "geodesic")
  }
  if (case == "example2") {
    tree_list <- replicate(n = sum(size), ape::rtree(n = 15, br = runif), 
                           simplify = FALSE)
    class(tree_list) <- "multiPhylo"
    x <- treespace::multiDist(tree_list, lambda = 0.5)
  }
  if (case == "example3") {
    x1 <- movMF::rmovMF(n = size[1], theta = 1 * c(0, 1, 1, 1, 1) / 2)
    x2 <- movMF::rmovMF(n = size[1], theta = 2 * c(2, 1, 1, 1, 1) / sqrt(8))
    x3 <- movMF::rmovMF(n = size[1], theta = 3 * c(4, 1, 1, 1, 1) / sqrt(20))
    x4 <- movMF::rmovMF(n = size[1], theta = 4 * c(6, 1, 1, 1, 1) / sqrt(40))
    x <- rbind(x1, x2, x3, x4)
    x <- Ball::nhdist(x, method = "geodesic")
  }
  if (case == "example4") {
    x1 <- movMF::rmovMF(n = sum(size[1]), theta = c(30, 30) *
                          rbind(c(0, 1),c(0, -1)), alpha = c(0.5, 0.5))
    x2 <- movMF::rmovMF(n = sum(size[2]), theta = c(35, 35) *
                          rbind(c(0, 1), c(0, -1)), alpha = c(0.5, 0.5))
    x3 <- movMF::rmovMF(n = sum(size[3]), theta = c(30, 30) *
                          rbind(c(1, 0), c(-1, 0)), alpha = c(0.5, 0.5))
    x4 <- movMF::rmovMF(n = sum(size[4]), theta = c(35, 35) *
                          rbind(c(1, 0), c(-1, 0)), alpha = c(0.5, 0.5))
    x <- rbind(x1, x2, x3, x4)
    x <- Ball::nhdist(x)
  }
  if (case == "example5") {
    x1 <- movMF::rmovMF(n = size[1], 
                        theta = c(30, 30) * rbind(c(1, 0, 0, 0), 
                                                  c(-1, 0, 0, 0)), 
                        alpha = c(0.5, 0.5))
    x2 <- movMF::rmovMF(n = size[2], 
                        theta = c(30, 30) * rbind(c(0, 1, 0, 0), 
                                                  c(0, -1, 0, 0)), 
                        alpha = c(0.5, 0.5))
    x3 <- movMF::rmovMF(n = size[3], 
                        theta = c(30, 30) * rbind(c(0, 0, 1, 0), 
                                                  c(0, 0, -1, 0)), 
                        alpha = c(0.5, 0.5))
    x4 <- movMF::rmovMF(n = size[4], 
                        theta = c(30, 30) * rbind(c(0, 0, 0, 1), 
                                                  c(0, 0, 0, -1)), 
                        alpha = c(0.5, 0.5))
    x <- rbind(x1, x2, x3, x4)
    x <- Ball::nhdist(x)
  }
  if (case == "example6") {
    tree1 <- replicate(n = size[1], 
                       ape::rtree(n = 15, br = runif, min = 0, max = 1),
                       simplify = FALSE)
    tree2 <- replicate(n = size[2], 
                       ape::rtree(n = 15, br = runif, min = 0, max = 0.5),
                       simplify = FALSE)
    tree3 <- replicate(n = size[3], 
                       ape::rtree(n = 15, br = runif, min = 0, max = 0.75),
                       simplify = FALSE)
    tree4 <- replicate(n = size[4], 
                       ape::rtree(n = 15, br = runif, min = 0, max = 0.25),
                       simplify = FALSE)
    tree_list <- c(tree1, tree2, tree3, tree4)
    class(tree_list) <- "multiPhylo"
    x <- treespace::multiDist(tree_list, lambda = 0.5)
  }
  if (case == "example7") {
    tree1 <- replicate(n = size[1], 
                       ape::rtree(n = 15, br = runif, min = 0, max = 12), 
                       simplify = FALSE)
    tree2 <- replicate(n = size[2], 
                       ape::rtree(n = 15, br = runif,min = 2, max = 10), 
                       simplify = FALSE)
    tree3 <- replicate(n = size[3], 
                       ape::rtree(n = 15, br = runif, min = 3, max = 8), 
                       simplify = FALSE)
    tree4 <- replicate(n = size[4], 
                       ape::rtree(n = 15, br = runif, min = 4, max = 6), 
                       simplify = FALSE)
    tree_list <- c(tree1, tree2, tree3, tree4)
    class(tree_list) <- "multiPhylo"
    x <- treespace::multiDist(tree_list, lambda = 0.5)
  }
  if (case == "example8") {
    tree1 <- replicate(n = size[1], 
                       ape::rtree(n = 15, br = rf, df1 = 2, df2 = 2),
                       simplify = FALSE)
    tree2 <- replicate(n = size[2], 
                       ape::rtree(n = 15, br = rf, df1 = 2, df2 = 3),
                       simplify = FALSE)
    tree3 <- replicate(n = size[3], 
                       ape::rtree(n = 15, br = rf, df1 = 2, df2 = 4),
                       simplify = FALSE)
    tree4 <- replicate(n = size[4], 
                       ape::rtree(n = 15, br = rf, df1 = 2, df2 = 5),
                       simplify = FALSE)
    tree_list <- c(tree1, tree2, tree3, tree4)
    class(tree_list) <- "multiPhylo"
    x <- treespace::multiDist(tree_list, lambda = 0.5)
  }
  if (case == "example9") {
    x1 <- movMF::rmovMF(n = size[1], theta = 3 * c(0, 1, 1, 1, 1) / 2)
    x2 <- movMF::rmovMF(n = sum(size[-1]), 
                        theta = 3 * c(2, 1, 1, 1, 1) / sqrt(8))
    x <- rbind(x1, x2)
    x <- Ball::nhdist(x)
  }
  if (case == "example10") {
    tree1 <- replicate(n = size[1], 
                       ape::rtree(n = 15, br = runif, min = 0, max = 0.5),
                       simplify = FALSE)
    tree2 <- replicate(n = sum(size[-1]), 
                       ape::rtree(n = 15, br = runif, min = 0, max = 1),
                       simplify = FALSE)
    tree_list <- c(tree1, tree2)
    class(tree_list) <- "multiPhylo"
    x <- treespace::multiDist(tree_list, lambda = 0.5)
  }
  list("x" = x)
}

testExecute <- function(seed, case = "example1", size = c(10, 10, 10, 10)) {
  dataset <- data_generator(seed = seed, case = case, size = size)
  set.seed(1); energy_pvalue <- energy::eqdist.etest(x = dataset[["x"]], 
                                                     sizes = size,
                                                     distance = TRUE, 
                                                     R = 399)$p.value
  bd_res <- Ball::bd.test(x = dataset[["x"]], 
                          size = size, num.permutations = 399,
                          seed = 1, distance = TRUE, num.threads = 0)
  bd_sum_pvalue <- bd_res$p.value
  bd_max_pvalue <- bd_res$complete.info$p.value['kbd.max.constant.pvalue']
  bd_maxsum_pvalue <- bd_res$complete.info$p.value['kbd.maxsum.constant.pvalue']
  #
  hhg_y <- mapply(function(x, y) {
    rep(x, y)
  }, 1:length(size), size, SIMPLIFY = FALSE)
  hhg_y <- unlist(hhg_y)
  hhg_dx <- as.matrix(dist(dataset[["x"]], diag = TRUE, upper = TRUE))
  set.seed(1); 
  hhg_pvalue <- HHG::hhg.test.k.sample(hhg_dx, hhg_y, nr.perm = 399, 
                                       nr.threads = 1)$perm.pval.hhg.sc
  c(energy_pvalue, hhg_pvalue, bd_sum_pvalue, bd_max_pvalue, bd_maxsum_pvalue)
}

simulation_wrapper <- function(seed, case = "example1", size_list =
                                 list(rep(10, 4), rep(20, 4), rep(30, 4),
                                      rep(40, 4), rep(50, 4))) {
  pvalue_list <- list()
  for(i in 1:length(size_list)) {
    pvalue_list[[i]] = testExecute(seed = seed, case = case, 
                                   size = size_list[[i]])
  }
  pvalue_list <- do.call("rbind", pvalue_list)
  pvalue_list < 0.05
}

calculate_power <- function(pvalue_mat_list){
  cols <- ncol(pvalue_mat_list[[1]])
  pvalue_mat_list <- lapply(1:cols, function(i){
    sapply(pvalue_mat_list,function(x) {x[,i]})
  })
  power_tab <- sapply(pvalue_mat_list, rowMeans)
  colnames(power_tab) <- c("energy", "hhg", "kbd_sum", "kbd_max", "kbd_maxsum")
  power_tab
}

sfInit(parallel = TRUE, cpus = 6)
sfLibrary(ape)
sfLibrary(treespace)
sfLibrary(movMF)
sfLibrary(energy)
sfLibrary(HHG)
sfLibrary(Ball)
sfExportAll()
replicate_num <- 500
example_1 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example1")
res1 <- t(calculate_power(example_1))
example_2 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example2")
res2 <- t(calculate_power(example_2))
example_3 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example3")
res3 <- t(calculate_power(example_3))
example_4 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example4")
res4 <- t(calculate_power(example_4))
example_5 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example5")
res5 <- t(calculate_power(example_5))
example_6 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example6")
res6 <- t(calculate_power(example_6))
example_7 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example7")
res7 <- t(calculate_power(example_7))
example_8 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example8")
res8 <- t(calculate_power(example_8))
example_9 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example9", 
                      size_list = list(rep(10, 10), rep(20, 10), rep(30, 10),
                                       rep(40, 10), rep(50, 10)))
res9 <- t(calculate_power(example_9))
example_10 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example10", 
                       size_list = list(rep(10, 10), rep(20, 10), rep(30, 10),
                                        rep(40, 10), rep(50, 10)))
res10 <- t(calculate_power(example_10))
sfStop()

###################### Visualize K-sample test result #####################
source("visualize-utilize.R")
bd_x_axis_break <- scale_x_continuous(breaks = c(10, 20, 30, 40, 50))

### Plot Figure 2 
p_res <- cbind.data.frame(rbind(res1, res2), 
                          example_list(1, 2, prefix = "5.1."),
                          kbd_method(nrep = 2))
colnames(p_res) <- bd_plot_column_list
p_res <- data_process(p_res)
y_axis <- "Type-I error rate"
y_limit <- ylim(c(0.00, 0.10))
p1 <- basic_plot(p_res) + labs(x = "Sample size of each group", y = y_axis) + 
  y_limit + bd_x_axis_break + type1error_intercept
p1
ggsave(filename = "../Figures/kbd_error.eps", plot = p1, width = 7, height = 4)

### Plot Figure 3
p_res <- cbind.data.frame(rbind(res3, res4, res5, res6, 
                                res7, res8, res9, res10), 
                          example_list(3, 10, prefix = "5.1."), 
                          kbd_method(nrep = 8))
colnames(p_res) <- bd_plot_column_list
p_res <- data_process(p_res)
y_axis <- "Empirical power"
y_limit <- ylim(c(0.0, 1.0))
p2 <- basic_plot(p_res) + labs(x = "Sample size of each group", y = y_axis) + 
  y_limit + bd_x_axis_break
p2
ggsave(filename = "../Figures/kbd_power.eps", plot = p2, 
       width = 7, height = 11.5)


######################### Test of mutual independence #########################
rm(list = ls()); gc(reset = TRUE)

### test of independence
generate_batch_spd <- function(param, p = 10)
{
  sampled_spd <- sapply(param, function(x) {
    clusterGeneration::genPositiveDefMat(dim = p, "eigen", lambdaLow = x[1],
                                         ratioLambda = x[2])[["Sigma"]]
  }, simplify = "array")
  symApply <- apply(sampled_spd, 3, Matrix::isSymmetric, 
                    tol = sqrt(.Machine$double.eps))
  if (all(symApply)) {
  } else {
    sampled_spd <- sapply(param, function(x) {
      clusterGeneration::genPositiveDefMat(dim = p, "eigen", lambdaLow = 1,
                                           ratioLambda = 10)[["Sigma"]]
    }, simplify = "array")
  }
  CovTools::CovDist(sampled_spd)
}

data_generator <- function(seed = 1, case = "example1", size = 30) {
  set.seed(seed)
  if(case == "example1") {
    x <- data.frame(t(data.frame(rep(1, size), rep(10, size))))
    dx <- generate_batch_spd(x)
    y <- data.frame(t(data.frame(rep(1, size), rep(10, size))))
    dy <- generate_batch_spd(y)
  }
  if(case == "example2") {
    fourier.base <- fda::create.fourier.basis(rangeval = c(0, 8 * pi), 
                                              nbasis = 3,
                                              period = 2 * pi)
    fourier.fd1 <- fda::fd(coef = matrix(runif(size * 3), nrow = 3),
                           basisobj = fourier.base)
    fourier.fdata1 <- t(fda::eval.fd(evalarg = seq(0, 8 * pi, length.out = 17),
                                     fdobj = fourier.fd1))
    fourier.fdata1 <- fda.usc::fdata(mdata = fourier.fdata1,
                                     argvals = seq(0, 8 * pi, length.out = 17))
    fourier.fd2 <- fda::fd(coef = matrix(runif(size * 3), nrow = 3),
                           basisobj = fourier.base)
    fourier.fdata2 <- t(fda::eval.fd(evalarg = seq(0, 8 * pi, length.out = 17),
                                     fdobj = fourier.fd2))
    fourier.fdata2 <- fda.usc::fdata(mdata = fourier.fdata2,
                                     argvals = seq(0, 8 * pi, length.out = 17))
    dx <- fda.usc::metric.lp(fourier.fdata1, lp = 0)
    dy <- fda.usc::metric.lp(fourier.fdata2, lp = 0)
  }
  if (case == "example3") {
    z1 <- runif(size, max = pi / 2)
    z2 <- rchisq(size, df = 1) + 1
    z3 <- z1 + rchisq(size, df = 1)
    z4 <- 1 + rchisq(size, df = 1)
    x <- data.frame(rbind(z1, z2))
    y <- data.frame(rbind(z3, z4))
    dx <- generate_batch_spd(x, p = 10)
    dy <- generate_batch_spd(y, p = 10)
  }
  if (case == "example4") {
    z1 <- runif(size, max = pi / 2)
    z2 <- rchisq(size, df = 1) + 1
    z3 <- 10 * abs(cos(3 * z1)) + rchisq(size, df = 1)
    z4 <- 1 + rchisq(size, df = 1)
    x <- data.frame(rbind(z1, z2))
    y <- data.frame(rbind(z3, z4))
    dx <- generate_batch_spd(x, p = 10)
    dy <- generate_batch_spd(y, p = 10)
  }
  if(case == "example5") {
    observe_point <- seq(0, 8 * pi, length.out = 17)
    fourier.base <- fda::create.fourier.basis(rangeval = c(0, 8 * pi), 
                                              nbasis = 3,
                                              period = 2 * pi)
    fourier.fd1 <- fda::fd(coef = matrix(rnorm(size * 3), nrow = 3), 
                           basisobj = fourier.base)
    fourier.fd2 <- fda::fd(coef = matrix(rnorm(size * 3), nrow = 3), 
                           basisobj = fourier.base)
    fourier.fd3 <- fda::fd(coef = matrix(rnorm(size * 3), nrow = 3), 
                           basisobj = fourier.base)
    fourier.fdata1 <- t(fda::eval.fd(evalarg = observe_point, 
                                     fdobj = fourier.fd1))
    fourier.fdata2 <- t(fda::eval.fd(evalarg = observe_point, 
                                     fdobj = fourier.fd2))
    fourier.fdata3 <- t(fda::eval.fd(evalarg = observe_point, 
                                     fdobj = fourier.fd3))
    x <- 4 * fourier.fdata1 * fourier.fdata2 * fourier.fdata3 +
      matrix(rnorm(n = size * ncol(fourier.fdata1)), nrow = size)
    y <- (fourier.fdata1 - fourier.fdata2)^2 + 2 * fourier.fdata3  +
      matrix(rnorm(n = size * ncol(fourier.fdata1)), nrow = size)
    x <- fda.usc::fdata(mdata = x, argvals = observe_point)
    y <- fda.usc::fdata(mdata = y, argvals = observe_point)
    dx <- fda.usc::metric.lp(x, lp = 0)
    dy <- fda.usc::metric.lp(y, lp = 0)
  }
  if(case == "example6") {
    observe_point <- seq(0, 8 * pi, length.out = 17)
    x <- rbinom(n = size, size = 1, prob = 0.5)
    yi <- rbinom(n = size, size = 1, prob = 0.5)
    ycoef <- matrix(nrow = size, ncol = 3)
    for (i in 1:size) {
      if (x[i] == 0 && yi[i] == 0) {
        ycoef[i, ] <- c(0, 0, 1)
      } else if (x[i] == 0 && yi[i] == 1) {
        ycoef[i, ] <- c(0, 0, -1)
      } else if (x[i] == 1 && yi[i] == 0) {
        ycoef[i, ] <- c(0, 1, 0)
      } else {
        ycoef[i, ] <- c(0, -1, 0)
      }
    }
    fourier_base <- fda::create.fourier.basis(rangeval = c(0, 8 * pi), 
                                              nbasis = 3,
                                              period = 2 * pi)
    fd <- fda::fd(coef = t(ycoef), basisobj = fourier_base) 
    fd_obs_data <- t(fda::eval.fd(evalarg = observe_point, fdobj = fd))
    fd_obs_data <- 10 * fd_obs_data + matrix(rnorm(n = size * 
                                                     ncol(fd_obs_data)),
                                             nrow = size)
    fd_obs_data <- fda.usc::fdata(mdata = fd_obs_data, argvals = observe_point)
    dy <- fda.usc::metric.lp(fd_obs_data, lp = 0)
    dx <- as.matrix(dist(x))
  }
  return(list("dx" = dx, "dy" = dy))
}

testExecute <- function(seed, case = "example1", size = 30) {
  dataset <- data_generator(seed = seed, case = case, size = size)
  ind_result <- Ball::bcov.test(x = dataset[["dx"]], y = dataset[["dy"]],
                                num.permutations = 399, seed = 1,
                                distance = TRUE, num.threads = 1)
  ind_result <- ind_result[["complete.info"]]
  bcov_pvalue <- ind_result[["p.value"]][1]
  bcov_prob_pvalue <- ind_result[["p.value"]][2]
  bcov_hhg_pvalue <- ind_result[["p.value"]][3]
  #
  set.seed(1); 
  hhg_pvalue <- HHG::hhg.test(Dx = as.matrix(dataset[["dx"]]),
                              Dy = as.matrix(dataset[["dy"]]), 
                              nr.perm = 399, 
                              nr.threads = 1)[["perm.pval.hhg.sc"]]
  set.seed(1); dcov_pvalue <- energy::dcov.test(x = as.dist(dataset[["dx"]]),
                                                y = as.dist(dataset[["dy"]]),
                                                R = 399)[["p.value"]]
  c(dcov_pvalue, hhg_pvalue, bcov_pvalue, bcov_prob_pvalue, bcov_hhg_pvalue)
}

simulation_wrapper <- function(seed, case = "example1", 
                               size_list = c(40, 60, 80, 100, 120))
{
  pvalue_list <- list()
  for(i in 1:length(size_list)) {
    pvalue_list[[i]] = testExecute(seed = seed, case = case, 
                                   size = size_list[i])
  }
  pvalue_list <- do.call("rbind", pvalue_list)
  pvalue_list < 0.05
}

calculate_power <- function(pvalue_mat_list){
  cols <- ncol(pvalue_mat_list[[1]])
  pvalue_mat_list <- lapply(1:cols, function(i){
    sapply(pvalue_mat_list, function(x) {x[,i]})
  })
  power_tab <- sapply(pvalue_mat_list, rowMeans)
  power_tab
}

sfInit(parallel = TRUE, cpus = 6)
sfLibrary(clusterGeneration)
sfLibrary(CovTools)
sfLibrary(fda)
sfLibrary(fda.usc)
sfLibrary(energy)
sfLibrary(HHG)
sfLibrary(Ball)
sfExportAll()
replicate_num <- 500
example_1 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example1")
res1 <- t(calculate_power(example_1))
example_2 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example2")
res2 <- t(calculate_power(example_2))
example_3 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example3")
res3 <- t(calculate_power(example_3))
example_4 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example4")
res4 <- t(calculate_power(example_4))
example_5 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example5")
res5 <- t(calculate_power(example_5))
example_6 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example6")
res6 <- t(calculate_power(example_6))

### test of mutual independence
double_normalized_distance <- function(dx) {
  size <- nrow(dx)
  C <- diag(size) - matrix(data = 1, nrow = size, ncol = size) / size
  -C %*% dx %*% C
}

distance_multivariate <- function(data_list, R = 199) {
  data_list <- lapply(data_list, double_normalized_distance)
  tm <- multivariance::total.multivariance(x = data_list)
  tm.resampled <- numeric(R)
  for (r in 1:R) {
    set.seed(r)
    tm.resampled[r] <- multivariance::total.multivariance(
      multivariance::sample.cdms(data_list))
  }
  dcov_pvalue <- (sum(tm.resampled > tm) + 1) / (R + 1)
  dcov_pvalue
}

data_generator <- function(seed = 1, case = "example7", size = 30) {
  set.seed(seed)
  observe_point <- seq(0, 8 * pi, length.out = 17)
  if(case == "example7") {
    z1 <- runif(size, max = pi / 2)
    z2 <- 1 + rchisq(size, df = 1)
    
    z3 <- z1 + EnvStats::rpareto(n = size, location = 0.8, shape = 1)
    z4 <- 1 + rchisq(size, df = 1)
    
    z5 <- z1 + EnvStats::rpareto(n = size, location = 0.8, shape = 1)
    z6 <- 1 + rchisq(size, df = 1)
    
    x <- data.frame(rbind(z1, z2))
    y <- data.frame(rbind(z3, z4))
    z <- data.frame(rbind(z5, z6))
    dx <- generate_batch_spd(x, p = 10)
    dy <- generate_batch_spd(y, p = 10)
    dz <- generate_batch_spd(z, p = 10)
  }
  if (case == "example8") {
    w1 <- rt(n = size, df = 1)
    w2 <- rt(n = size, df = 1)
    x <- abs(w1) + 1
    y <- exp(8 * sin(w1 - w2)) + 1
    z <- 1 + (w2 - w1)^2
    x <- data.frame(t(data.frame(1, x)))
    y <- data.frame(t(data.frame(1, y)))
    z <- data.frame(t(data.frame(1, z)))
    dx <- generate_batch_spd(x, p = 10)
    dy <- generate_batch_spd(y, p = 10)
    dz <- generate_batch_spd(z, p = 10)
  }
  if (case == "example9") {
    fourier_base <- fda::create.fourier.basis(rangeval = c(0, 8 * pi), 
                                              nbasis = 3,
                                              period = 2 * pi)
    
    fd1 <- fda::fd(coef = replicate(n = size, expr = c(0, 1, 0)), 
                   basisobj = fourier_base) 
    fd2 <- fda::fd(coef = replicate(n = size, expr = c(0, 0, 1)), 
                   basisobj = fourier_base) 
    fourier.fdata1 <- t(fda::eval.fd(evalarg = observe_point, fdobj = fd1))
    fourier.fdata2 <- t(fda::eval.fd(evalarg = observe_point, fdobj = fd2))
    
    z1 <- rbinom(n = size, size = 1, prob = 0.5)
    z2 <- rbinom(n = size, size = 1, prob = 0.5)
    x <- z1
    y <- 1 - z2
    z <- ifelse(z1 == z2, 1, 0)
    fz <- fy <- fx <- fourier.fdata1
    for (i in 1:size) {
      fx[i, ] <- ifelse(x[i] == 1, fourier.fdata1[i, ], fourier.fdata2[i, ])
      fy[i, ] <- ifelse(y[i] == 1, fourier.fdata1[i, ], fourier.fdata2[i, ])
      fz[i, ] <- ifelse(z[i] == 1, fourier.fdata1[i, ], fourier.fdata2[i, ])
    }
    fx <- fda.usc::fdata(mdata = fx, argvals = observe_point)
    fy <- fda.usc::fdata(mdata = fy, argvals = observe_point)
    fz <- fda.usc::fdata(mdata = fz, argvals = observe_point)
    dx <- fda.usc::metric.lp(fx, lp = 0)
    dy <- fda.usc::metric.lp(fy, lp = 0)
    dz <- fda.usc::metric.lp(fz, lp = 0)
  }
  if (case == "example10") {
    x <- rbinom(n = size, size = 1, prob = 0.5)
    yi <- rbinom(n = size, size = 1, prob = 0.5)
    ycoef <- matrix(nrow = size, ncol = 3)
    for (i in 1:size) {
      if (x[i] == 0 && yi[i] == 0) {
        ycoef[i, ] <- c(0, 0, 1)
      } else if (x[i] == 0 && yi[i] == 1) {
        ycoef[i, ] <- c(0, 0, -1)
      } else if (x[i] == 1 && yi[i] == 0) {
        ycoef[i, ] <- c(0, 1, 0)
      } else {
        ycoef[i, ] <- c(0, -1, 0)
      }
    }
    fourier_base <- fda::create.fourier.basis(rangeval = c(0, 8 * pi), 
                                              nbasis = 3,
                                              period = 2 * pi)
    fd <- fda::fd(coef = t(ycoef), basisobj = fourier_base) 
    fd_obs_data <- t(fda::eval.fd(evalarg = seq(0, 8 * pi, length.out = 17), 
                                  fdobj = fd))
    fd_obs_data <- 10 * fd_obs_data + matrix(rnorm(n = size * 17), nrow = size)
    fd_obs_data <- fda.usc::fdata(fd_obs_data, argvals = observe_point)
    noise <- fda.usc::fdata(mdata = matrix(rnorm(n = size * 17), nrow = size),
                            argvals = observe_point)
    
    dx <- as.matrix(dist(x))
    dy <- fda.usc::metric.lp(fd_obs_data, lp = 0)
    dz <- fda.usc::metric.lp(noise, lp = 0)
  }
  return(list("dx" = dx, "dy" = dy, "dz" = dz))
}

testExecute <- function(seed, case = "example1", size = 30) {
  dataset <- data_generator(seed = seed, case = case, size = size)
  ind_result <- Ball::bcov.test(x = dataset, num.permutations = 399, seed = 1, 
                                distance = TRUE, num.threads = 1)
  ind_result <- ind_result[["complete.info"]]
  bcov_pvalue <- ind_result[["p.value"]][1]
  bcov_prob_pvalue <- ind_result[["p.value"]][2]
  bcov_hhg_pvalue <- ind_result[["p.value"]][3]
  set.seed(1); dcov_pvalue <- distance_multivariate(dataset, R = 399)
  c(dcov_pvalue, bcov_pvalue, bcov_prob_pvalue, bcov_hhg_pvalue)
}

sfLibrary(multivariance)
sfExport("testExecute", "data_generator", "distance_multivariate",
         "double_normalized_distance")
example_7 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example7")
res7 <- t(calculate_power(example_7))
example_8 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example8")
res8 <- t(calculate_power(example_8))
example_9 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example9")
res9 <- t(calculate_power(example_9))
example_10 <- sfLapply(1:replicate_num, simulation_wrapper, case = "example10")
res10 <- t(calculate_power(example_10))
sfStop()

################# Visualize test of mutual independence test result #################

source("visualize-utilize.R")
bcov_x_axis_break <- scale_x_continuous(breaks = c(40, 60, 80, 100, 120))

### Plot Figure 4
p_res <- cbind.data.frame(rbind(res1, res2), 
                          example_list(1, 2, prefix = "5.2."),
                          bcov_method(nrep = 2))
colnames(p_res) <- bcov_plot_column_list
p_res <- data_process(p_res)
y_axis <- "Type-I error rate"
y_limit <- ylim(c(0.00, 0.10))
p3 <- basic_plot(p_res) + labs(x = "Sample size", y = y_axis) + 
  y_limit + bcov_x_axis_break + type1error_intercept
p3
ggsave(filename = "../Figures/bcov_error.eps", plot = p3, 
       width = 7, height = 4)

### Plot Figure 5
p_res <- cbind.data.frame(rbind(res3, res4, res5, res6), 
                          example_list(3, 6, prefix = "5.2."),
                          bcov_method(nrep = 4))
colnames(p_res) <- bcov_plot_column_list
p_res <- data_process(p_res)
y_axis <- "Empirical power"
y_limit <- ylim(c(0.0, 1.0))
p4 <- basic_plot(p_res) + labs(x = "Sample size", y = y_axis) + 
  y_limit + bcov_x_axis_break
p4
ggsave(filename = "../Figures/bcov_power.eps", plot = p4, 
       width = 7, height = 6.5)

### Plot Figure 6
p_res <- cbind.data.frame(rbind(res7, res8, res9, res10), 
                          example_list(7, 10, prefix = "5.2.", rep_num = 4),
                          kbcov_method(nrep = 4))
colnames(p_res) <- kbcov_plot_column_list
p_res <- data_process(p_res)
y_axis <- "Empirical power"
y_limit <- ylim(c(0.0, 1.0))
p5 <- basic_plot(p_res) + labs(x = "Sample size", y = y_axis) + 
  y_limit + bcov_x_axis_break + 
  guides(color = guide_legend(nrow = 2, byrow = TRUE))
p5
ggsave(filename = "../Figures/kbcov_power.eps", plot = p5, 
       width = 7, height = 6.5)

############################## Runtime analysis  ##############################
rm(list = setdiff(ls(), c("generate_batch_spd", "double_normalized_distance",
                          "distance_multivariate")))
gc(reset = TRUE)

cbPalette <- c("#e41a1c","#e6ab02","#4daf4a","#984ea3","#ff7f00","#377eb8")
basic_plot <- function(dat) {
  ggplot(dat, mapping = aes(x = size, y = Runtime, color = Package)) +
    geom_point() + geom_line() +
    theme_bw() + 
    theme(legend.position = c(0.3, 0.78),
          text = element_text(color = "black"),
          axis.title = element_text(face = "bold", size = rel(0.8)),
          axis.text = element_text(size = rel(1), color = "black"),
          legend.title = element_text(face = "bold"),
          legend.text = element_text(face = "bold"),
          legend.background = element_rect(fill = "transparent"),
          legend.key.size = unit(0.8, "lines")
    ) + scale_colour_manual(values = cbPalette)
}

kbd_runtime <- function(i, size) {
  x <- movMF::rmovMF(n = sum(size), 30 * c(1, 0, 0))
  dx <- Ball::nhdist(x, method = "geodesic")
  k <- 4
  sub_size <- size / k
  size_vec <- rep(sub_size, k)
  times_evaluate <- microbenchmark::microbenchmark(
    "energy" = { 
      set.seed(1); 
      ed_res <- energy::eqdist.etest(x = dx, 
                                     sizes = size_vec,
                                     distance = TRUE, 
                                     R = 399) }, 
    "HHG" = { 
      set.seed(1);
      hhg_pvalue <- HHG::hhg.test.k.sample(Dx = dx,
                                           y = rep(1:k, each = sub_size),
                                           nr.perm = 399, nr.threads = 1)
      },
    "Ball1" = { 
      bd_res1 <- Ball::bd.test(x = dx, size = size_vec, num.permutations = 399,
                               seed = 1, distance = TRUE, num.threads = 1) 
      }, 
    "Ball2" = { 
      bd_res1 <- Ball::bd.test(x = dx, size = size_vec, num.permutations = 399,
                               seed = 1, distance = TRUE, num.threads = 2) 
      }, 
    "Ball4" = { 
      bd_res1 <- Ball::bd.test(x = dx, size = size_vec, num.permutations = 399,
                               seed = 1, distance = TRUE, num.threads = 4) 
      }, 
    "Ball8" = { 
      bd_res1 <- Ball::bd.test(x = dx, size = size_vec, num.permutations = 399,
                               seed = 1, distance = TRUE, num.threads = 8) 
      }, 
    times = 1)
  index <- c(which(times_evaluate[["expr"]] == "energy"),
             which(times_evaluate[["expr"]] == "HHG"), 
             which(times_evaluate[["expr"]] == "Ball1"),
             which(times_evaluate[["expr"]] == "Ball2"),
             which(times_evaluate[["expr"]] == "Ball4"),
             which(times_evaluate[["expr"]] == "Ball8"))
  times_evaluate[index, "time"]
}

bcov_runtime <- function(i, size) {
  set.seed(i)
  x <- data.frame(t(data.frame(rep(1, size), rep(10, size))))
  y <- data.frame(t(data.frame(rep(1, size), rep(10, size))))
  dx <- generate_batch_spd(x)
  dy <- generate_batch_spd(y)
  times_evaluate <- microbenchmark::microbenchmark(
    "energy" = { 
      set.seed(1); ed_res <- energy::dcov.test(x = dx, y = dy, R = 399) 
      }, 
    "HHG" = { 
      set.seed(1); 
      hhg_res <- HHG::hhg.test(Dx = as.matrix(dx), Dy = as.matrix(dy),
                               nr.perm = 399, nr.threads = 1) 
      },
    "Ball1" = { 
      bd_res1 <- Ball::bcov.test(x = dx, y = dy, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 1) }, 
    "Ball2" = { 
      bd_res1 <- Ball::bcov.test(x = dx, y = dy, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 2) }, 
    "Ball4" = { 
      bd_res1 <- Ball::bcov.test(x = dx, y = dy, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 4) }, 
    "Ball8" = { 
      bd_res1 <- Ball::bcov.test(x = dx, y = dy, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 8) }, 
    times = 1)
  index <- c(which(times_evaluate[["expr"]] == "energy"),
             which(times_evaluate[["expr"]] == "HHG"), 
             which(times_evaluate[["expr"]] == "Ball1"),
             which(times_evaluate[["expr"]] == "Ball2"),
             which(times_evaluate[["expr"]] == "Ball4"),
             which(times_evaluate[["expr"]] == "Ball8"))
  times_evaluate[index, "time"]
}

kbcov_runtime <- function(i, size) {
  set.seed(i)
  w1 <- rt(n = size, df = 1)
  w2 <- rt(n = size, df = 1)
  x <- abs(w1) + 1
  y <- exp(8 * sin(w1 - w2)) + 1
  z <- 1 + 0.5 * (w2 - w1)^2
  x <- data.frame(t(data.frame(1, x)))
  y <- data.frame(t(data.frame(1, y)))
  z <- data.frame(t(data.frame(1, z)))
  dx <- generate_batch_spd(x, p = 10)
  dy <- generate_batch_spd(y, p = 10)
  dz <- generate_batch_spd(z, p = 10)
  dataset <- list(dx, dy, dz)
  times_evaluate <- microbenchmark::microbenchmark(
    "energy" = { 
      set.seed(1); ed_res <- distance_multivariate(dataset, R = 399) 
      }, 
    "Ball1" = { 
      bd_res1 <- Ball::bcov.test(x = dataset, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 1) 
      }, 
    "Ball2" = { 
      bd_res1 <- Ball::bcov.test(x = dataset, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 2) 
      }, 
    "Ball4" = { 
      bd_res1 <- Ball::bcov.test(x = dataset, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 4) 
      }, 
    "Ball8" = { 
      bd_res1 <- Ball::bcov.test(x = dataset, num.permutations = 399,
                                 seed = 1, distance = TRUE, num.threads = 8) 
      }, 
    times = 1)
  index <- c(which(times_evaluate[["expr"]] == "energy"), 
             which(times_evaluate[["expr"]] == "Ball1"),
             which(times_evaluate[["expr"]] == "Ball2"),
             which(times_evaluate[["expr"]] == "Ball4"),
             which(times_evaluate[["expr"]] == "Ball8"))
  times_evaluate[index, "time"]
}

rep_num <- 20
seed_vec <- rep(1:rep_num, 5)
diff_size_list <- c(40, 80, 120, 160, 200)
diff_size_vec <- rep(diff_size_list, each = rep_num)
dep_size_list <- c(40, 60, 80, 100, 120)
dep_size_vec <- rep(dep_size_list, each = rep_num)

kbd_runtime_result <- mapply(kbd_runtime, size = diff_size_vec, i = seed_vec)
kbd_runtime_summary <- sapply(diff_size_list, function(size_value) {
  rowMeans(kbd_runtime_result[, diff_size_vec == size_value, drop = FALSE])
}) / 10^9
rownames(kbd_runtime_summary) <- c("energy", "HHG", 
                                   "Ball (1 thread)", "Ball (2 threads)",
                                   "Ball (4 threads)", "Ball (8 threads)")
colnames(kbd_runtime_summary) <- diff_size_list / 4

bcov_runtime_result <- mapply(bcov_runtime, size = dep_size_vec, i = seed_vec)
bcov_runtime_summary <- sapply(dep_size_list, function(size_value) {
  rowMeans(bcov_runtime_result[, dep_size_vec == size_value, drop = FALSE])
}) / 10^9
rownames(bcov_runtime_summary) <- c("energy", "HHG", 
                                    "Ball (1 thread)", "Ball (2 threads)",
                                    "Ball (4 threads)", "Ball (8 threads)")
colnames(bcov_runtime_summary) <- dep_size_list

kbcov_runtime_result <- mapply(kbcov_runtime, size = dep_size_vec, i = seed_vec)
kbcov_runtime_summary <- sapply(dep_size_list, function(size_value) {
  rowMeans(kbcov_runtime_result[, dep_size_vec == size_value, drop = FALSE])
}) / 10^9
rownames(kbcov_runtime_summary) <- c("multivariance", 
                                     "Ball (1 thread)", "Ball (2 threads)",
                                     "Ball (4 threads)", "Ball (8 threads)")
colnames(kbcov_runtime_summary) <- dep_size_list

###################### Visualize runtimes analysis result ######################
library("grid")
library("gridExtra")

### Plot Figure 7
p_kbd <- melt(kbd_runtime_summary, varnames = c("Package", "size"), 
              value.name = "Runtime")
p1 <- basic_plot(p_kbd)
p1 <- p1 + 
  scale_y_continuous(breaks = seq(0, 0.88, length.out = 5),
                     limits = c(0, 0.88)) +
  xlab(label = "Sample size of each group")
p_bcov <- melt(bcov_runtime_summary, varnames = c("Package", "size"), 
               value.name = "Runtime")
p2 <- basic_plot(p_bcov)
p2 <- p2 + 
  scale_y_continuous(breaks = seq(0, 1.48, length.out = 5),
                     limits = c(0, 1.48)) +
  xlab(label = "Sample size")
p_kbcov <- melt(kbcov_runtime_summary, varnames = c("Package", "size"), 
                value.name = "Runtime")
p3 <- basic_plot(p_kbcov)
p3 <- p3 + xlab(label = "Sample size") + 
  theme(legend.position = c(0.3, 0.81)) + 
  scale_y_continuous(breaks = seq(0, 6, length.out = 5), limits = c(0, 6)) +
  scale_colour_manual(values = cbPalette[-2])

# pdf("../Figures/runtime.pdf", height = 3.5, width = 9)
grid.arrange(arrangeGrob(p1, p2, p3, ncol = 3, nrow = 1))
x <- c(0.01, 0.342, 0.673)
y <- 0.95
grid.text(c("A", "B", "C"),
          x = c(x[1], x[2], x[3]),
          y = rep(y, 3),
          just = c("left", "bottom"),
          gp = gpar(fontsize = 14))
# dev.off()

# save.image("JSS.RData")
